# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import os

import bpy
import rna_prop_ui

from ._vendor import utils

addon = utils.bpy.addon.Addon()


# VALID = (
#     "NodeSocketString",
#     "NodeSocketBool",
#     "NodeSocketMaterial",
#     "NodeSocketVector",
#     "NodeSocketInt",
#     "NodeSocketMenu",
#     "NodeSocketCollection",
#     "NodeSocketGeometry",
#     "NodeSocketTexture",
#     "NodeSocketFloat",
#     "NodeSocketColor",
#     "NodeSocketObject",
#     "NodeSocketRotation",
#     "NodeSocketMatrix",
#     "NodeSocketImage",
# )


PRIMITIVE_TYPES = (
    "NodeSocketBool",
    "NodeSocketColor",
    "NodeSocketFloat",
    "NodeSocketInt",
    "NodeSocketRotation",
    "NodeSocketString",
    "NodeSocketVector",
)


def snake_case(s):
    return "".join(["_" + c.lower() if c.isupper() else c for c in s]).lstrip("_")


def snake_upper(s):
    if s.isupper():
        return s
    else:
        return snake_case(s).upper()


def is_primitive_socket(tree_item):
    return (
        hasattr(tree_item, "default_value") and tree_item.socket_type in PRIMITIVE_TYPES
    )


def get_treeitem_path(tree_item):
    if tree_item.parent.name == "":
        return tree_item.name
    else:
        return "%s::%s" % (tree_item.parent.name, tree_item.name)


def get_interface_inputs(node_tree):
    inputs_index = 0
    for tree_item in node_tree.interface.items_tree:
        if tree_item.item_type == "SOCKET" and tree_item.in_out == "INPUT":
            yield inputs_index, tree_item
            inputs_index += 1


def iter_supported_sockets(node):
    node_sockets = node.inputs

    for index, tree_item in get_interface_inputs(node.node_tree):
        if is_primitive_socket(tree_item):
            socket = node_sockets[index]
            if not socket.hide_value:
                name = get_treeitem_path(tree_item)
                yield socket, name, tree_item


def iter_missing_sockets(node, preset):
    for socket, name, tree_item in iter_supported_sockets(node):
        if name not in preset.inputs:
            yield socket, name, tree_item


def iter_editable_sockets(node, preset):
    for socket, name, tree_item in iter_supported_sockets(node):
        if name in preset.inputs:
            preset_item = preset.inputs[name]

            if is_convertable_socket_type(
                tree_item.socket_type, preset_item.socket_type
            ):
                yield preset_item, socket, tree_item


def check_sockets_uniqueness(node):
    socket_list = [name for _, name, _ in iter_supported_sockets(node)]
    socket_set = set(socket_list)
    return len(socket_list) == len(socket_set)


def get_supported_sockets(node):
    sockets = {}
    for socket, name, tree_item in iter_supported_sockets(node):
        sockets[socket] = (name, tree_item)
    return sockets


def get_missing_sockets(node, preset):
    sockets = {}
    for socket, name, tree_item in iter_missing_sockets(node, preset):
        sockets[socket] = (name, tree_item)
    return sockets


def get_editable_sockets(node, preset):
    sockets = {}
    for preset_item, socket, tree_item in iter_editable_sockets(node, preset):
        sockets[preset_item] = (socket, tree_item)
    return sockets


def sort_preset_sockets(node, preset):
    editable_sockets = get_editable_sockets(node, preset)

    indexed_preset = [item.name for item in preset.inputs]
    indexed_sockets = []
    for item in editable_sockets:
        if item.name in indexed_preset:
            indexed_sockets.append(item.name)

    for item in editable_sockets:
        old_index = indexed_preset.index(item.name)
        new_index = indexed_sockets.index(item.name)
        if new_index != old_index:
            preset.inputs.move(old_index, new_index)
            indexed_preset.insert(new_index, indexed_preset.pop(old_index))


def is_convertable_socket_type(socket1, socket2):
    # NOTE: NodeSocket subtypes values are internally the same has subvalues.abs
    ## No need to handle subtypes differently. This compares "bl_idname"
    ## up until a common root to only distinguishe base type
    common = os.path.commonprefix([socket1, socket2])

    if common != "NodeSocket":
        return True
    elif socket1.startswith(
        ("NodeSocketInt", "NodeSocketFloat"),
    ) and socket2.startswith(
        ("NodeSocketInt", "NodeSocketFloat"),
    ):
        return True
    else:
        return False


def commit_preset_item_value(preset_item, init_value, property_args):
    preset_item["value"] = init_value
    preset_item.id_properties_ensure()
    property = preset_item.id_properties_ui("value")
    property.update(**property_args)


def set_preset_item_value(preset_item, value, socket_type):
    preset_item["value"] = value
    preset_item.socket_type = socket_type


def get_property_args_from_tree_item(tree_item):
    property_args = {}

    socket_type = tree_item.bl_socket_idname
    match socket_type:
        case "NodeSocketColor":
            init_value = [0.0, 0.0, 0.0, 0.0]
            property_args["subtype"] = "COLOR"
        case "NodeSocketVector":
            init_value = [0.0, 0.0, 0.0]
        case "NodeSocketRotation":
            init_value = [0.0, 0.0, 0.0]
            property_args["subtype"] = "EULER"
        case "NodeSocketFloat":
            init_value = 0.0
        case "NodeSocketInt":
            init_value = 0
        case "NodeSocketString":
            init_value = ""
        case "NodeSocketBool":
            init_value = False
        case _:
            raise NotImplementedError(
                "Socket of type :%s not supported yet" % socket_type
            )

    if hasattr(tree_item, "subtype"):
        property_args["subtype"] = tree_item.subtype

    if hasattr(tree_item, "max_value"):
        property_args["max"] = property_args["soft_max"] = tree_item.max_value
        property_args["min"] = property_args["soft_min"] = tree_item.min_value

    if isinstance(tree_item.default_value, bpy.types.bpy_prop_array):
        default_value = list(tree_item.default_value)
    else:
        default_value = None
    property_args["default"] = default_value

    return init_value, property_args


def init_preset_item_value(preset_item, value, socket_type, tree_item):
    # https://docs.blender.org/api/current/bpy_types_enum_items/node_socket_data_type_items.html
    # https://docs.blender.org/api/current/bpy_types_enum_items/property_subtype_items.html
    init_value, property_args = get_property_args_from_tree_item(tree_item)

    commit_preset_item_value(preset_item, init_value, property_args)
    set_preset_item_value(preset_item, value, socket_type)


def create_unique_preset(rna_id, name):
    name = utils.bpy.find_next_available_name(rna_id, name)
    preset = rna_id.add()
    preset.name = name
    return preset


class NODE_OT_Group_AddPresetItem(bpy.types.Operator):
    """Adds a preset item"""

    bl_idname = "nodegroup.add_preset_item"
    bl_label = "Add Preset Item"
    bl_options = {"UNDO"}

    index: bpy.props.IntProperty(name="Selected Preset")
    buffer = []

    def get_buffer(self, context):
        # NOTE: ENUM_FLAG has a maximum length of 32
        ## Python int to C long fails at length of 31
        return __class__.buffer[:30]

    def update_buffer(self, context):
        buffer = []
        i = 0

        for socket, socket_info in self.sockets.items():
            name, tree_item = socket_info
            if i > 30:
                i = 0
            bit_flag = 2**i
            buffer.append(
                (
                    name,
                    name,
                    "",
                    bit_flag,
                )
            )
            i += 1

        __class__.buffer = buffer

    selection: bpy.props.EnumProperty(
        items=get_buffer,
        options={"ENUM_FLAG"},
    )

    def draw(self, context):
        layout = self.layout

        box = layout.box().row()
        box.column().props_enum(self, "selection")

    def invoke(self, context, event):
        node = context.active_node
        preset = node.node_tree.group_presets.presets[self.index]
        self.sockets = get_missing_sockets(node, preset)
        self.update_buffer(context)
        return context.window_manager.invoke_props_dialog(self)

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    group_presets = node.node_tree.group_presets
                    preset = group_presets.presets[group_presets.active_index]
                    if preset is not None:
                        return len(get_missing_sockets(node, preset)) > 0
        return False

    def execute(self, context):
        if not self.selection:
            return {"CANCELLED"}

        node = context.active_node
        preset = node.node_tree.group_presets.presets[self.index]

        for socket, socket_info in self.sockets.items():
            name, tree_item = socket_info

            if name not in self.selection:
                continue

            item = preset.inputs.add()
            item.name = name
            init_preset_item_value(
                item, socket.default_value, socket.bl_idname, tree_item
            )

        sort_preset_sockets(node, preset)
        context.area.tag_redraw()
        return {"FINISHED"}


class NODE_OT_Group_DeletePresetItem(bpy.types.Operator):
    """Delete the chosen preset item"""

    bl_idname = "nodegroup.remove_preset_item"
    bl_label = "Delete Preset Item"
    bl_options = {"UNDO"}

    index: bpy.props.IntProperty(name="Selected Preset Item")

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    group_presets = node.node_tree.group_presets
                    return group_presets.presets[group_presets.active_index] is not None
        return False

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets
        preset = group_presets.presets[group_presets.active_index]
        preset.inputs.remove(self.index)
        return {"FINISHED"}


class NODE_OT_Group_SortPresetItems(bpy.types.Operator):
    """Sorts the chosen preset"""

    bl_idname = "nodegroup.sort_preset_items"
    bl_label = "Sort Preset Items"
    bl_options = {"UNDO"}

    index: bpy.props.IntProperty(name="Selected Preset Item")

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    group_presets = node.node_tree.group_presets
                    return group_presets.presets[group_presets.active_index] is not None
        return False

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets
        preset = group_presets.presets[group_presets.active_index]
        sort_preset_sockets(node, preset)
        return {"FINISHED"}


class NODE_OT_Group_ClearPresetItems(bpy.types.Operator):
    """Cleares the chosen preset"""

    bl_idname = "nodegroup.clear_preset_items"
    bl_label = "Clear Preset Items"
    bl_options = {"UNDO"}

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    group_presets = node.node_tree.group_presets
                    return group_presets.presets[group_presets.active_index] is not None
        return False

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets
        active_preset = group_presets.presets[group_presets.active_index]
        active_preset.inputs.clear()
        return {"FINISHED"}


class NODE_OT_Group_SavePreset(bpy.types.Operator):
    """Saves the current settings as a preset"""

    bl_idname = "nodegroup.save_preset"
    bl_label = "Save Preset"
    bl_options = {"UNDO"}

    partial: bpy.props.BoolProperty()
    buffer = []
    sockets = {}

    def get_buffer(self, context):
        # NOTE: ENUM_FLAG has a maximum length of 32
        ## Python int to C long fails at length of 31
        return __class__.buffer[:30]

    def get_buffer2(self, context):
        return __class__.buffer[31:62]

    @staticmethod
    def get_buffer_len():
        return len(__class__.buffer)

    def update_buffer(self, context):
        buffer = []
        i = 0
        for socket, socket_info in self.sockets.items():
            name, tree_item = socket_info
            if i > 30:
                i = 0
            bit_flag = 2**i
            buffer.append(
                (
                    name,
                    name,
                    "",
                    bit_flag,
                )
            )
            i += 1

        __class__.buffer = buffer

    selection: bpy.props.EnumProperty(
        items=get_buffer,
        options={"ENUM_FLAG"},
    )

    selection2: bpy.props.EnumProperty(
        items=get_buffer2,
        options={"ENUM_FLAG"},
    )

    def draw(self, context):
        layout = self.layout

        box = layout.box().row()
        box.column().props_enum(self, "selection")

        if self.get_buffer_len() > 32:
            box.column().props_enum(self, "selection2")

    def invoke(self, context, event):
        self.sockets = get_supported_sockets(context.active_node)

        if event.shift:
            self.update_buffer(context)
            self.partial = True
            return context.window_manager.invoke_props_dialog(self)
        else:
            self.partial = False
            return self.execute(context)

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                return node.bl_idname.endswith("NodeGroup")
        return False

    def execute(self, context):
        selection = self.selection | self.selection2
        if self.partial and not selection:
            return {"CANCELLED"}

        session = addon.session
        node = context.active_node
        group_presets = node.node_tree.group_presets
        presets = group_presets.presets

        name = session.preset_name
        preset = create_unique_preset(presets, name)

        for socket, socket_info in self.sockets.items():
            name, tree_item = socket_info

            if self.partial and name not in selection:
                continue

            if socket.hide_value:
                continue

            item = preset.inputs.add()
            item.name = name
            init_preset_item_value(
                item, socket.default_value, socket.bl_idname, tree_item
            )

        group_presets.active_index = presets.find(preset.name)
        context.area.tag_redraw()
        return {"FINISHED"}


class NODE_OT_Group_DuplicatePreset(bpy.types.Operator):
    """Duplicate the chosen preset"""

    bl_idname = "nodegroup.duplicate_preset"
    bl_label = "Duplicate Preset"
    bl_options = {"UNDO"}

    index: bpy.props.IntProperty(name="Selected Preset")

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    return len(node.node_tree.group_presets.presets) > 0

    def invoke(self, context, event):
        node = context.active_node
        preset = node.node_tree.group_presets.presets[self.index]
        self.sockets = get_editable_sockets(node, preset)
        return self.execute(context)

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets

        old_preset = group_presets.presets[self.index]
        preset = create_unique_preset(group_presets.presets, old_preset.name)

        for old_item, socket_tup in self.sockets.items():
            socket, tree_item = socket_tup

            item = preset.inputs.add()
            item.name = old_item.name
            init_preset_item_value(item, old_item["value"], socket.bl_idname, tree_item)
        sort_preset_sockets(node, preset)

        group_presets.active_index = group_presets.presets.find(preset.name)
        return {"FINISHED"}


class NODE_OT_Group_UpdatePreset(bpy.types.Operator):
    """Update the chosen preset with the current node state"""

    bl_idname = "nodegroup.update_preset"
    bl_label = "Update Preset"
    bl_options = {"UNDO"}

    index: bpy.props.IntProperty(name="Selected Preset")

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    return len(node.node_tree.group_presets.presets) > 0

    def invoke(self, context, event):
        node = context.active_node
        preset = node.node_tree.group_presets.presets[self.index]
        self.sockets = get_editable_sockets(node, preset)
        return self.execute(context)

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets

        preset = group_presets.presets[self.index]

        for item, socket_tup in self.sockets.items():
            socket, tree_item = socket_tup
            init_preset_item_value(
                item, socket.default_value, socket.bl_idname, tree_item
            )

        sort_preset_sockets(node, preset)
        return {"FINISHED"}


class NODE_OT_Group_ApplyPreset(bpy.types.Operator):
    """Applies the chosen preset to the node"""

    bl_idname = "nodegroup.apply_preset"
    bl_label = "Apply Preset"
    bl_options = {"UNDO"}

    multi: bpy.props.BoolProperty()
    index: bpy.props.IntProperty(name="Selected Preset")

    partial: bpy.props.BoolProperty()
    buffer = []
    sockets = {}

    def get_buffer(self, context):
        # NOTE: ENUM_FLAG has a maximum length of 32
        ## Python int to C long fails at length of 31
        return __class__.buffer[:30]

    @staticmethod
    def get_buffer_len():
        return len(__class__.buffer)

    def update_buffer(self, context):
        buffer = []
        i = 0

        for item in self.sockets:
            if i > 30:
                i = 0
            bit_flag = 2**i
            buffer.append((item.name, item.name, "", bit_flag))
            i += 1

        __class__.buffer = buffer

    selection: bpy.props.EnumProperty(
        items=get_buffer,
        options={"ENUM_FLAG"},
    )

    def draw(self, context):
        layout = self.layout

        box = layout.box().row()
        box.column().props_enum(self, "selection")

    def invoke(self, context, event):
        self.multi = event.alt

        node = context.active_node
        preset = node.node_tree.group_presets.presets[self.index]
        self.sockets = get_editable_sockets(node, preset)

        if event.shift:
            self.update_buffer(context)
            self.partial = True
            return context.window_manager.invoke_props_dialog(self)
        else:
            self.partial = False
            return self.execute(context)

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    return len(node.node_tree.group_presets.presets) > 0
        return False

    def execute(self, context):
        if self.partial and not self.selection:
            return {"CANCELLED"}

        active_node = context.active_node
        if not self.multi:
            nodes = [active_node]
        else:
            nodes = []
            for node in context.selected_nodes:
                print(node.node_tree == active_node.node_tree)
                if (
                    node.bl_idname.endswith("NodeGroup")
                    and node.node_tree == active_node.node_tree
                ):
                    nodes.append(node)

        preset = active_node.node_tree.group_presets.presets[self.index]
        group_presets = node.node_tree.group_presets
        for node in nodes:
            sockets = get_editable_sockets(node, preset)
            for item in sockets:
                if self.partial and item.name not in self.selection:
                    continue

                socket, _ = sockets[item]
                socket.default_value = item["value"]

            group_presets.active_index = group_presets.presets.find(preset.name)
            self.index = -1
        return {"FINISHED"}


class NODE_OT_Group_DeletePreset(bpy.types.Operator):
    """Delete the chosen preset"""

    bl_idname = "nodegroup.remove_preset"
    bl_label = "Delete Preset"
    bl_options = {"UNDO"}

    index: bpy.props.IntProperty(name="Selected Preset")

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    return len(node.node_tree.group_presets.presets) > 0
        return False

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets
        presets = group_presets.presets

        presets.remove(self.index)

        if self.index <= group_presets.active_index:
            if len(presets) == 0:
                group_presets.active_index = 0
            else:
                group_presets.active_index -= 1
        return {"FINISHED"}


class NODE_OT_Group_ClearPresets(bpy.types.Operator):
    """Removes all pretes"""

    bl_idname = "nodegroup.clear_presets"
    bl_label = "Clear All Presets"
    bl_options = {"UNDO"}

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    return len(node.node_tree.group_presets.presets) > 0
        return False

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets
        group_presets.presets.clear()
        group_presets.active_index = 0
        return {"FINISHED"}


class NODE_OT_Group_ClipboardCopyPresets(bpy.types.Operator):
    """Copies presets to clipboard"""

    bl_idname = "nodegroup.clipboard_copy_presets"
    bl_label = "Copy"
    bl_options = {"UNDO"}

    partial = bpy.props.BoolProperty()
    buffer = []

    def get_buffer(self, context):
        # NOTE: ENUM_FLAG has a maximum length of 32
        ## Python int to C long fails at length of 31
        return __class__.buffer[:30]

    @staticmethod
    def get_buffer_len():
        return len(__class__.buffer)

    def update_buffer(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets

        buffer = []
        i = 0
        for preset in group_presets.presets:
            if preset.inputs != "":
                if i > 30:
                    i = 0
                bit_flag = 2**i
                buffer.append((preset.name, preset.name, "", bit_flag))
                i += 1

        __class__.buffer = buffer

    selection: bpy.props.EnumProperty(
        items=get_buffer,
        options={"ENUM_FLAG"},
    )

    def draw(self, context):
        layout = self.layout

        box = layout.box().row()
        box.column().props_enum(self, "selection")

    def invoke(self, context, event):
        if event.shift:
            self.update_buffer(context)
            self.partial = True
            return context.window_manager.invoke_props_dialog(self)
        else:
            self.partial = False
            return self.execute(context)

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                if node.bl_idname.endswith("NodeGroup"):
                    return len(node.node_tree.group_presets.presets) > 0
        return False

    def execute(self, context):
        if self.partial and not self.selection:
            return {"CANCELLED"}

        node = context.active_node
        group_presets = node.node_tree.group_presets

        options = utils.bpy.rna2json.ExportOptions()
        options.include_custom_props = True

        presets = utils.bpy.rna2json.dump(
            attr=(group_presets, "presets"), output="RAW", options=options
        )

        if self.partial:
            _temp = []
            selection = self.selection

            filtered_data = []
            for preset in presets:
                if preset["name"] in selection:
                    filtered_data.append(preset)

            presets = filtered_data

        context.window_manager.clipboard = utils.bpy.rna2json.dump(raw=presets)
        return {"INTERFACE"}


class NODE_OT_Group_ClipboardPastePresets(bpy.types.Operator):
    """Pastes presets from clipboard"""

    bl_idname = "nodegroup.clipboard_paste_presets"
    bl_label = "Paste"
    bl_options = {"UNDO"}

    @classmethod
    def poll(cls, context):
        if context.area.type == "NODE_EDITOR":
            node = context.active_node
            if node is not None:
                return node.bl_idname.endswith("NodeGroup")
        return False

    def execute(self, context):
        node = context.active_node
        group_presets = node.node_tree.group_presets
        presets = group_presets.presets

        try:
            clipboard_data = utils.bpy.rna2json.load(
                context.window_manager.clipboard,
                to_dict=True,
            )
        except Exception:
            self.report({"WARNING"}, "Invalid JSON, aborting.")
            return {"CANCELLED"}

        for py_preset in clipboard_data:
            # NOTE: Skip empty presets
            if "inputs" in py_preset:
                name = py_preset.pop("name")
                preset = presets.get(name)
                if preset is not None:
                    i = 1
                    result = 1
                    while result >= 0:
                        _name = "%s.%03d" % (name, i)
                        result = presets.find(_name)
                        i += 1
                    name = _name

                preset = presets.add()
                preset.name = name

                options = utils.bpy.rna2json.ImportOptions()
                options.include_custom_props = True
                utils.bpy.rna2json.load(
                    raw=py_preset["inputs"], to_attr=(preset, "inputs"), options=options
                )

        context.area.tag_redraw()
        return {"FINISHED"}


class NODE_UL_Group_PresetList(bpy.types.UIList):
    def draw_item(
        self, context, layout, data, item, icon, active_data, active_propname, index
    ):
        active_index = data.active_index
        layout = layout.row(align=True)

        utils.bpy.ui.empty(
            layout,
            enabled=False,
            icon="RESTRICT_SELECT_OFF"
            if index == active_index
            else "RESTRICT_SELECT_ON",
        )

        layout.operator(
            NODE_OT_Group_ApplyPreset.bl_idname,
            text=item.name,
            emboss=False,
        ).index = index

        layout.operator(
            NODE_OT_Group_DeletePreset.bl_idname,
            text="",
            icon="REMOVE",
            emboss=False,
        ).index = index

    def draw_filter(self, context, layout):
        layout = layout.column()

        row = layout.row(align=True)
        row.prop(self, "filter_name", text="", icon="VIEWZOOM")
        row.prop(self, "use_filter_invert", text="", icon="ARROW_LEFTRIGHT")

        row.separator()
        row.prop(self, "use_filter_sort_reverse", text="", icon="SORT_ASC")

    def filter_items(self, context, data, propname):
        """Filter and order items in the list"""
        helper_funcs = bpy.types.UI_UL_list

        filtered = []
        ordered = []
        items = [item for item in getattr(data, propname)]

        filtered = helper_funcs.filter_items_by_name(
            self.filter_name, self.bitflag_filter_item, items, "name"
        )

        ordered = helper_funcs.sort_items_by_name(items, "name")
        return filtered, ordered


class NODE_PT_Group_Presets(bpy.types.Panel):
    bl_space_type = "NODE_EDITOR"
    bl_region_type = "UI"
    bl_category = "Node"
    bl_label = "Presets"

    @classmethod
    def poll(cls, context):
        node = context.active_node
        if node is not None:
            if node.bl_idname.endswith("NodeGroup"):
                return node.node_tree is not None
        return False

    def draw_header_preset(self, context):
        layout = self.layout
        buttons = layout.row(align=True)
        buttons.operator(
            NODE_OT_Group_ClipboardCopyPresets.bl_idname,
            text="",
            icon="COPYDOWN",
            emboss=False,
        )
        buttons.operator(
            NODE_OT_Group_ClipboardPastePresets.bl_idname,
            text="",
            icon="PASTEDOWN",
            emboss=False,
        )

    def draw(self, context):
        session = addon.session
        node = context.active_node
        ntree = node.node_tree
        group_presets = ntree.group_presets

        layout = self.layout.column(align=True)

        header = layout.box().column()
        info = header.row()

        if not check_sockets_uniqueness(node):
            utils.bpy.ui.alert(info, text="All sockets need to be unique.")
        else:
            info.prop(session, "preset_name", text="")

            buttons = info.row(align=True)
            buttons.alignment = "RIGHT"

            buttons.operator(
                NODE_OT_Group_SavePreset.bl_idname,
                text="",
                icon="ADD",
                emboss=False,
            )

            buttons.operator(
                NODE_OT_Group_DuplicatePreset.bl_idname,
                icon="DUPLICATE",
                text="",
                emboss=False,
            ).index = group_presets.active_index

            buttons.operator(
                NODE_OT_Group_ClearPresets.bl_idname,
                text="",
                icon="TRASH",
                emboss=False,
            )

            layout.template_list(
                "NODE_UL_Group_PresetList",
                "Preset List",
                group_presets,
                "presets",
                group_presets,
                "active_index",
            )

            if len(group_presets.presets) > group_presets.active_index:
                active_index = group_presets.active_index
                preset = group_presets.presets[group_presets.active_index]
                layout = self.layout.column(align=True)
                header = layout.box().row()

                header.prop(preset, "name", text="")
                buttons = header.row(align=True)

                buttons.operator(
                    NODE_OT_Group_SortPresetItems.bl_idname,
                    text="",
                    icon="LINENUMBERS_ON",
                    emboss=False,
                )

                buttons.operator(
                    NODE_OT_Group_UpdatePreset.bl_idname,
                    text="",
                    icon="IMPORT",
                    emboss=False,
                ).index = active_index

                buttons.operator(
                    NODE_OT_Group_ClearPresetItems.bl_idname,
                    text="",
                    icon="TRASH",
                    emboss=False,
                )

                body = layout.box().column()
                if len(preset.inputs):
                    editable_sockets = get_editable_sockets(node, preset)

                    for i, item in enumerate(preset.inputs):
                        row = body.row()

                        label, content = utils.bpy.ui.split2(
                            row,
                            label_alignment="EXPAND",
                            factor=0.5,
                        )
                        label.alert = item not in editable_sockets

                        label.prop(item, "name", text="")

                        content.column().prop(
                            item, rna_prop_ui.rna_idprop_quote_path("value"), text=""
                        )

                        buttons = content.row()
                        buttons.alignment = "RIGHT"

                        buttons.operator(
                            NODE_OT_Group_DeletePresetItem.bl_idname,
                            text="",
                            icon="PANEL_CLOSE",
                            emboss=False,
                        ).index = i

                row = body.row()
                label, content = utils.bpy.ui.split2(
                    row,
                    label_alignment="EXPAND",
                )

                content.alignment = "RIGHT"
                content.operator(
                    NODE_OT_Group_AddPresetItem.bl_idname,
                    icon="ADD",
                    text="",
                    emboss=False,
                ).index = active_index


class NodeGroupPresetSocket(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty(name="Socket Name", default="Socket")
    socket_type: bpy.props.StringProperty(name="Socket Type", default="")


class NodeGroupPreset(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty(name="Preset Name", default="Preset")
    inputs: bpy.props.CollectionProperty(type=NodeGroupPresetSocket)


class NodeTree_GroupPresets(bpy.types.PropertyGroup):
    active_index: bpy.props.IntProperty(
        name="Active PresetList Index",
    )
    presets: bpy.props.CollectionProperty(
        name="Presets Collection",
        type=NodeGroupPreset,
    )


@addon.property
class NGPRESETS_WindowManager(bpy.types.PropertyGroup):
    preset_name: bpy.props.StringProperty(name="New Preset Name", default="Preset")


CLASSES = [
    NODE_OT_Group_AddPresetItem,
    NODE_OT_Group_DeletePresetItem,
    NODE_OT_Group_ClearPresetItems,
    NODE_OT_Group_SortPresetItems,
    NODE_OT_Group_SavePreset,
    NODE_OT_Group_DuplicatePreset,
    NODE_OT_Group_UpdatePreset,
    NODE_OT_Group_ApplyPreset,
    NODE_OT_Group_DeletePreset,
    NODE_OT_Group_ClearPresets,
    NODE_OT_Group_ClipboardCopyPresets,
    NODE_OT_Group_ClipboardPastePresets,
    NODE_UL_Group_PresetList,
    NODE_PT_Group_Presets,
    NodeGroupPresetSocket,
    NodeGroupPreset,
    NodeTree_GroupPresets,
    NGPRESETS_WindowManager,
]

PROPS = [
    NGPRESETS_WindowManager,
]


def register():
    utils.bpy.register_classes(CLASSES)
    addon.set_properties(PROPS)

    setattr(
        bpy.types.NodeTree,
        "group_presets",
        bpy.props.PointerProperty(type=NodeTree_GroupPresets),
    )


def unregister():
    utils.bpy.unregister_classes(CLASSES)
