# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import bpy
import inspect

from .. import std

properties = {}
keymaps = []


class _Caller:
    def __init__(self):
        frame = inspect.currentframe()
        try:
            self.frame = frame.f_back.f_back
            self.mod = inspect.getmodule(self.frame)
            self.package = self.mod.__name__
            self.path = self.package.split(".")

            self.name = self.path[-1]

            self.addon_name = self.path[2]
            self.addon_package = ".".join(self.path[:3])

            self.internal_package = ".".join(self.path[3:])
            self.internal_path = self.path[3:]

            self.parent_internal_package = ".".join(self.path[3:-1])
            self.parent_internal_path = self.path[3:-1]
        finally:
            del frame


# CREDIT: https://blender.stackexchange.com/questions/227358#227402
class Addon:
    def __init__(self, **kwargs):
        self.caller = _Caller()
        self.local = self.caller.mod

        self.name = self.caller.addon_name
        self.package = self.caller.addon_package

        self.properties = properties

    @property
    def are_preferences_registered(self):
        if self.package in bpy.context.preferences.addons:
            return bpy.context.preferences.addons[self.package].preferences is not None
        return False

    @property
    def root_preferences(self):
        """bpy.context.preferences.addons[demo].preferences"""
        return bpy.context.preferences.addons[self.package].preferences

    @property
    def parent_preferences(self):
        return std.rgetattr(
            bpy.context.preferences.addons[self.package].preferences,
            self.caller.parent_internal_path,
        )

    @property
    def preferences(self):
        return std.rgetattr(
            bpy.context.preferences.addons[self.package].preferences,
            self.caller.internal_path,
        )

    @property
    def root_session(self, root=False):
        """bpy.context.window_manager.demo"""
        return getattr(bpy.context.window_manager, self.name, None)

    @property
    def session(self, root=False):
        return std.rgetattr(
            bpy.context.window_manager, [self.name, *self.caller.internal_path]
        )

    @property
    def keyconfig(self):
        return bpy.context.window_manager.keyconfigs.addon

    @staticmethod
    def keymap_append(km_tuple):
        keymaps.append(km_tuple)

    @staticmethod
    def remove_keymaps():
        for km, idname in keymaps:
            index = km.keymap_items.find(idname)
            if index > -1:
                kmi = km.keymap_items[index]
                km.keymap_items.remove(kmi)

    # CREDIT: https://github.com/nutti/Screencast-Keys/commit/51925cbe31045eeeb7283c88b6c01d3ef1df2c2d
    def property(self, cls):
        """Generates PropertyGroup from child modules (if any). And adds itself
        to parent key in {properties} dictionary.
        """
        parent_module = ".".join([self.name, *self.caller.parent_internal_path])
        module = ".".join([self.name, *self.caller.internal_path])

        prefix, name = cls.__name__.split("_", 1)
        if not prefix.isupper():
            raise ValueError(
                "Invalid class name: %s\nClass name is required to have a prefix like 'DEMO_ClassName'"
                % cls.__name__
            )

        kind = name.split("_", 1)[0]

        if kind not in properties:
            properties.update({kind: {}})

        # Generate child modules' reference
        if module in properties[kind]:
            for child in properties[kind][module]:
                attr = child.__module__.split(".")[-1]
                cls.__annotations__[attr] = bpy.props.PointerProperty(type=child)

        # Add itself to parent key in dictionary
        if parent_module not in properties[kind]:
            properties[kind].update({parent_module: []})

        if parent_module != module:
            properties[kind][parent_module].append(cls)
        return cls

    def set_property(self, cls):
        prefix, name = cls.__name__.split("_", 1)
        if not prefix.isupper():
            raise ValueError(
                "Invalid class name: %s\nClass name is required to have a prefix like 'DEMO_ClassName'"
                % cls.__name__
            )

        kind = name.split("_")[0]

        if kind != "Preferences":
            object = getattr(bpy.types, kind)
            setattr(object, self.name, bpy.props.PointerProperty(type=cls))

    def set_properties(self, props):
        for cls in props:
            self.set_property(cls)

    def get_property(self, kind, root=False, parent=False):
        object = getattr(bpy.context, kind)
        if root:
            return getattr(object, self.name, None)
        if parent:
            return std.rgetattr(object, [self.name, *self.caller.parent_internal_path])
        else:
            return std.rgetattr(object, [self.name, *self.caller.internal_path])

    def del_property(self, cls):
        prefix, name = cls.__name__.split("_", 1)
        if not prefix.isupper():
            raise ValueError(
                "Invalid class name: %s\nClass name is required to have a prefix like 'DEMO_ClassName'"
                % cls.__name__
            )

        kind = name.split("_")[0]

        if kind != "Preferences":
            object = getattr(bpy.types, kind)
            delattr(object, self.name)

    def del_properties(self, props):
        for cls in reversed(props):
            self.del_property(cls)
