# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import inspect
import importlib

import bpy

from .. import std


def register_classes(classes):
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister_classes(classes):
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)


def iter_submodules(module=None):
    if module is None:
        frame = inspect.currentframe().f_back
        module = inspect.getmodule(frame)

    properties_tup = None
    for name, submodule in std.iter_submodules(module):
        # NOTE: Yield properties last to be safe regarding registration
        if name != "properties":
            yield name, submodule
        else:
            properties_tup = (name, submodule)

    if properties_tup is not None:
        yield properties_tup[0], properties_tup[1]


def get_submodules(module=None):
    if module is None:
        frame = inspect.currentframe().f_back
        module = inspect.getmodule(frame)
    return {name: submodule for name, submodule in iter_submodules(module)}


def register_submodules(module=None):
    if module is None:
        frame = inspect.currentframe().f_back
        module = inspect.getmodule(frame)

    if hasattr(module, "__submodules__"):
        submodules = module.__submodules__
    else:
        submodules = get_submodules(module)

    register_modules(submodules.values())


def unregister_submodules(module=None):
    if module is None:
        frame = inspect.currentframe().f_back
        module = inspect.getmodule(frame)

    if hasattr(module, "__submodules__"):
        submodules = module.__submodules__
    else:
        submodules = get_submodules(module)

    unregister_modules(submodules.values())


def register_modules(modules):
    for submodule in modules:
        if hasattr(submodule, "register"):
            submodule.register()


def unregister_modules(modules):
    for submodule in modules:
        if hasattr(submodule, "unregister"):
            submodule.unregister()


def enum_from_list(item_list, raw=False):
    if raw:
        return [(str(item), str(item).capitalize(), "") for item in item_list]
    else:
        return [(str(item).upper(), str(item).capitalize(), "") for item in item_list]


def validate_properties(dictionary):
    return all(
        [
            std.rgetattr(bpy, attribute.split(".")) == value
            for attribute, value in dictionary.items()
        ]
    )


def get_data_path(object, key=None):
    parent = repr(object.id_data)

    if key is not None:
        child = std.rgetattr(object, key.split("."))
    else:
        child = object

    return "%s.%s" % (parent, child.path_from_id())


def find_next_available_name(datablock, base_name):
    i = 1
    new_name = base_name
    result = datablock.find(base_name)
    while result != -1:
        new_name = "%s.%03d" % (base_name, i)
        i += 1
        result = datablock.find(new_name)
    return new_name
