# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import warnings

import bpy


def get_type_table(t):
    """Returns the type table of a given id type."""
    # DOC: https://docs.blender.org/api/current/bpy.types.ID.html
    # DOC: https://docs.blender.org/api/current/bpy_types_enum_items/id_type_items.html
    key = ""
    match t:
        case bpy.types.Object | "OBJECT":
            key = "objects"
        case bpy.types.Collection | "COLLECTION":
            key = "collections"
        case bpy.types.NodeTree | "NODETREE":
            key = "node_groups"
        case bpy.types.Material | "MATERIAL":
            key = "materials"
        case _:
            warnings.warn("IDType not supported: %s" % t)
    return key, getattr(bpy.data, key)


def obt(datablock, name, scope=None, force=False, overwrite=None, **kwargs):
    """Returns an object given a scope and an id, optionally creates it if does not already exist"""
    if scope is None:
        scope = datablock
    elif len(scope) > 0 and not isinstance(scope[0], str):
        scope = [item.name for item in scope]

    if name in scope:
        match overwrite:
            case "NEW":
                bldat = _create(datablock, name, **kwargs)
            case "SOFT":
                datablock[name].name = f"{name}.old"
                bldat = _create(datablock, name, **kwargs)
            case "HARD":
                old = datablock[name]
                old.name = f"{name}.###"
                bldat = _create(datablock, name, **kwargs)
                old.user_remap(bldat)
                datablock.remove(old)
            case None:
                bldat = datablock[name]
    elif force:
        if name in datablock:
            datablock[name].name = f"{name}.old"
        bldat = _create(datablock, name, **kwargs)
    else:
        bldat = None
    return bldat


def _create(datablock, name, **kwargs):
    match datablock:
        case bpy.data.objects | bpy.data.lights:
            bldat = datablock.new(name, kwargs["data"])
        case bpy.data.node_groups:
            bldat = datablock.new(name, kwargs["type"])
        case _:
            match type(datablock.bl_rna):
                case bpy.types.Nodes:
                    bldat = datablock.new(kwargs["type"])
                    bldat.name = name
                case _:
                    bldat = datablock.new(name)
    return bldat


class Preview:
    """Shadow class of ImagePreview used to copy from one instance to another"""

    # DOC: https://docs.blender.org/api/current/bpy.types.ImagePreview.html
    image_size = None
    image_pixels = None

    def __init__(self, image_size, image_pixels):
        self.image_size = image_size
        self.image_pixels = image_pixels


def copy_preview(id):
    preview_size = [*id.preview.image_size]

    preview_buffer = [None] * len(id.preview.image_pixels)
    id.preview.image_pixels.foreach_get(preview_buffer)

    return Preview(preview_size, preview_buffer)


def add_preview(id, preview):
    id.preview_ensure()
    id.preview.image_size = preview.image_size
    id.preview.image_pixels.foreach_set(preview.image_pixels)


def remove_preview(id):
    id.preview.image_size = [0, 0]
