# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import dataclasses
import json

from . import export_rna
from . import import_rna
from . import constructors
from . import typemap
from ... import std


@dataclasses.dataclass
class ExportOptions:
    ignore_default = False
    include_custom_props = False


@dataclasses.dataclass
class ImportOptions:
    include_custom_props = False


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def dump(
    rna_path=None,
    attr=None,
    raw=None,
    output="JSON",
    file_path=None,
    attr_filter={},
    options=None,
):
    if options is None:
        options = ExportOptions()

    typemap.build(attr_filter)

    if rna_path is not None:
        root = export_rna.parse_path(options, rna_path)
    elif attr is not None:
        obj = attr[0]
        key = attr[1]
        root = export_rna.parse_prop(options, {}, obj, key, None)
        root = root[key]
    elif raw is not None:
        root = raw
    else:
        raise ValueError("Invalid source. Use at least one from [rna_path, atttr, raw]")

    match output:
        case "RAW":
            return root
        case "JSON":
            return json.dumps(root, cls=JSONEncoder)
        case "FILE":
            json_str = json.dumps(root, cls=JSONEncoder)
            with open(file_path, "w") as file:
                file.write(json_str)
            return file_path
        case _:
            raise ValueError("Output Mode: %s" % output)


def load(
    json_str=None,
    raw=None,
    file_path=None,
    to_dict=False,
    to_attr=None,
    custom_constructors={},
    options=None,
):
    if options is None:
        options = ImportOptions()

    if json_str is not None:
        root = json.loads(json_str)
    elif raw is not None:
        root = raw
    elif file_path is not None:
        with open(file_path, "r") as file:
            root = json.load(file)
    else:
        raise ValueError("Invalid source. Either use json_str, raw or file_path")

    if to_dict:
        return root
    elif to_attr is not None:
        obj = to_attr[0]
        key = to_attr[1]
        root = {key: root}

        constructors.build(custom_constructors)
        import_rna.parse_prop(options, root, obj, key)
        return None
    else:
        constructors.build(custom_constructors)
        import_rna.parse_root(options, root)
        return None
