# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import bpy

from ... import std

attrmap = {}

included_classes = tuple()
excluded_classes = tuple()

_INCLUDE_CLASSES = ()
_EXCLUDE_CLASSES = (
    bpy.types.Operator,
    bpy.types.Menu,
    bpy.types.Panel,
    bpy.types.KeyingSet,
    bpy.types.Header,
)

TYPEMAP = {
    bpy.types.Preferences: {
        "exclude": [
            "active_section",
            "is_dirty",
        ],
    },
    bpy.types.ID: {
        "exclude": [
            "animation_data",
            "asset_data",
            "bl_description",
            "bl_icon",
            "bl_label",
            "grease_pencil",
            "is_runtime_data",
            "library_weak_reference",
            "library",
            "override_library",
            "preview",
            "tag",
            "use_extra_user",
            "use_fake_user",
        ],
    },
    bpy.types.Node: {
        "exclude": [
            "bl_description",
            "bl_label",
            "bl_icon",
            "bl_static_type",
            "bl_width_default",
            "bl_width_max",
            "bl_width_min",
            "bl_height_default",
            "bl_height_max",
            "bl_height_min",
            "internal_links",
            "width",
            "height",
        ],
        "default": {
            "show_options": True,
            "color": "#9b9b9b",
        },
    },
    bpy.types.NodeSocket: {
        "force": [
            "identifier",
        ],
        "exclude": [
            "bl_label",
            "link_limit",
            "name",
            "type",
        ],
        "default": {
            "enabled": True,
        },
    },
    bpy.types.NodeSocketVector: {"default": {"default_value": [0, 0, 0]}},
}


def _add_prop(properties, prop, prop_id, forced_props, props_default):
    if not prop.is_skip_save:
        if (
            prop_id in forced_props
            or not prop.is_readonly
            or type(prop)
            in (
                bpy.types.PointerProperty,
                bpy.types.CollectionProperty,
            )
        ):
            properties.update({prop_id: props_default.get(prop_id, None)})


def build(
    attr_filter={},
    include_cls=tuple(),
    exclude_cls=tuple(),
):
    included_classes = _INCLUDE_CLASSES + include_cls
    excluded_classes = _EXCLUDE_CLASSES + exclude_cls

    typemap = TYPEMAP.copy()
    for t in attr_filter:
        to_add = attr_filter[t]

        if t not in typemap:
            _map = typemap[t] = {}
        else:
            _map = typemap[t]

        if "force" in to_add:
            if "force" not in _map:
                _map["force"] = to_add["force"]
            else:
                _map["force"].extend(to_add["force"])

        if "include" in to_add:
            if "include" not in _map:
                _map["include"] = to_add["include"]
            else:
                _map["include"].extend(to_add["include"])

        if "exclude" in to_add:
            if "exclude" not in _map:
                _map["exclude"] = to_add["exclude"]
            else:
                _map["exclude"].extend(to_add["exclude"])

        if "default" in to_add:
            if "default" not in _map:
                _map["default"] = to_add["default"]
            else:
                _map["default"] = dict(
                    std.mergedicts(_map["default"], to_add["default"])
                )

    bpy_classes = {
        getattr(bpy.types, attr) for attr in dir(bpy.types) if not attr.startswith("_")
    }
    arg_classes = {t for t in attr_filter.keys()}
    all_classes = bpy_classes.union(arg_classes)

    for cls in all_classes:
        if issubclass(cls, excluded_classes):
            continue

        if len(included_classes) > 0 and not issubclass(cls, included_classes):
            continue

        bl_rna = getattr(cls, "bl_rna", None)
        # Needed to skip classes added to the modules `__dict__`.
        if bl_rna is None:
            continue

        attr_type = type(bl_rna)

        properties = {}

        forced_props = []
        included_props = []
        excluded_props = []
        props_default = {}

        if attr_type in typemap:
            _map = typemap[attr_type]
            if "force" in _map:
                forced_props.extend(_map["force"])
            if "include" in _map:
                included_props.extend(_map["include"])
            if "exclude" in _map:
                excluded_props.extend(_map["exclude"])
            if "default" in _map:
                props_default = _map["default"]

        for t in type(bl_rna).__mro__:
            if t in typemap:
                _map = typemap[t]
                if "force" in _map:
                    forced_props.extend(_map["force"])
                if "include" in _map:
                    included_props.extend(_map["include"])
                if "exclude" in _map:
                    excluded_props.extend(_map["exclude"])
                if "default" in _map:
                    props_default = dict(std.mergedicts(_map["default"], props_default))

        if len(included_props) > 0:
            rna_keys = bl_rna.properties.keys()
            for prop_id in included_props:
                if prop_id in rna_keys:
                    prop = bl_rna.properties[prop_id]
                    _add_prop(properties, prop, prop_id, forced_props, props_default)
        elif len(excluded_props) > 0:
            for prop_id, prop in bl_rna.properties.items():
                if prop_id not in excluded_props:
                    _add_prop(properties, prop, prop_id, forced_props, props_default)
        else:
            for prop_id, prop in bl_rna.properties.items():
                _add_prop(properties, prop, prop_id, forced_props, props_default)

        if "rna_type" in properties:
            del properties["rna_type"]
        attrmap[attr_type] = properties


def map_group(attr):
    bl_rna = getattr(attr, "bl_rna", None)
    # Needed to skip classes added to the modules `__dict__`.
    if bl_rna is None:
        return None

    # # to support skip-save we can't get all props
    properties = bl_rna.properties.keys()
    properties = {}
    for prop_id, prop in bl_rna.properties.items():
        if not prop.is_skip_save:
            if not prop.is_readonly or type(prop) in (
                bpy.types.PointerProperty,
                bpy.types.CollectionProperty,
            ):
                properties.update({prop_id: None})

        if "rna_type" in properties:
            del properties["rna_type"]
    return properties
