# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import functools
import inspect
import importlib
import itertools
import string
import types


def iter_submodules(module=None):
    if module is None:
        frame = inspect.currentframe().f_back
        module = inspect.getmodule(frame)

    pkg_name = module.__name__

    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if isinstance(attr, types.ModuleType):
            attr_pkg = attr.__name__
            if attr_pkg.rsplit(".", 1)[0] == pkg_name:
                yield attr_name, attr


def deepreload(module=None):
    if module is None:
        frame = inspect.currentframe().f_back
        module = inspect.getmodule(frame)

    def _reload(module):
        for name, submodule in iter_submodules(module):
            _reload(submodule)
            setattr(module, name, importlib.reload(submodule))

    _reload(module)


def rgetattr(object, names):
    for name in names:
        object = getattr(object, name)
    return object


def rsetattr(object, names, value):
    last = names.pop()
    rgetattr(object, names)
    setattr(object, last, value)


def traverse_tree(tree):
    yield tree
    for child in tree.children:
        yield from traverse_tree(child)


def pad_lists(*lists, padding=None):
    padded = [[] for _ in lists]
    for lst in itertools.zip_longest(*lists, fillvalue=padding):
        for i, elem in enumerate(lst):
            padded[i].append(elem)
    return padded


def mergedicts(dict1, dict2):
    for k in set(dict1) | set(dict2):
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield k, dict(mergedicts(dict1[k], dict2[k]))
            else:
                # If one of the values is not a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                yield (k, dict2[k])
                # Alternatively, replace this with exception raiser to alert you of value conflicts
        elif k in dict1:
            yield k, dict1[k]
        else:
            yield k, dict2[k]


# CREDIT: https://stackoverflow.com/a/48984697
def num2alpha(num):
    def div_mode_base26(n):
        a, b = divmod(n, 26)
        if b == 0:
            return a - 1, b + 26
        return a, b

    chars = []
    while num > 0:
        num, d = div_mode_base26(num)
        chars.append(string.ascii_uppercase[d - 1])
    return "".join(reversed(chars))


def alpha2num(chars):
    return functools.reduce(
        lambda r, x: r * 26 + x + 1, map(string.ascii_uppercase.index, chars), 0
    )
