# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import bpy
import idprop

from . import typemap


def parse_custom_prop(options, node, data, prop):
    subdata = data[prop]
    subnode = {}

    property = data.id_properties_ui(prop)
    subnode.update(property.as_dict())

    match type(subdata):
        case idprop.types.IDPropertyArray:
            value = subdata.to_list()
        case idprop.types.IDPropertyGroup:
            value = subdata.to_dict()
        case _:
            value = subdata

    subnode["value"] = value
    node[prop] = subnode
    return node


def parse_prop(options, node, data, prop, parent, default=None):
    def items2node(options, prop, subdata, subdata_fixed, data):
        subdata_node = {}
        if not issubclass(subdata_fixed, typemap.excluded_classes):
            parse_node(options, subdata_node, subdata, data)
        return subdata_node

    if data is None:
        return node

    subdata = getattr(data, prop)
    if subdata == data:
        return node

    if subdata is None:
        return node

    if subdata == parent:
        return node

    rna_props = data.bl_rna.properties
    prop_rna = rna_props[prop]
    if options.ignore_default and subdata == getattr(prop_rna, "default", None):
        return node

    subvalue = None
    prop_type = type(prop_rna.rna_type)
    match prop_type:
        case bpy.types.BoolProperty:
            subvalue = subdata
        case bpy.types.StringProperty:
            subvalue = subdata
        case bpy.types.FloatProperty | bpy.types.IntProperty:
            if prop_rna.is_array:
                subvalue = []
                for i in range(prop_rna.array_length):
                    subvalue.append(round(subdata[i], 5))
            else:
                subvalue = round(subdata, 5)
        case bpy.types.EnumProperty:
            if prop_rna.is_enum_flag:
                subvalue = [v for v in subdata]
            else:
                subvalue = subdata
        case bpy.types.CollectionProperty:
            if len(subdata) > 0:
                subvalue = []
                for subdata_item in subdata:
                    if subdata_item is not None:
                        item_node = parse_node(options, {}, subdata_item, data)
                        subvalue.append(item_node)
        case bpy.types.PointerProperty:
            subdata_fixed = type(prop_rna.fixed_type)
            subdata_node = items2node(
                options,
                prop,
                subdata,
                subdata_fixed,
                data,
            )
            if len(subdata_node) > 0:
                subvalue = subdata_node

    if subvalue is not None:
        if options.ignore_default or subvalue != default:
            node[prop] = subvalue
    return node


def parse_node(options, node, data, parent):
    data_type = type(data)

    if issubclass(data_type, typemap.excluded_classes):
        return None

    # NOTE: point-cache has eternal nested pointer to itself.
    if data == parent:
        return None

    if data_type not in typemap.attrmap:
        try:
            typemap.attrmap[data_type] = typemap.map_group(data)
        except Exception:
            import traceback

            print(traceback.format_exc())
            print(data_type, data, parent, node)
            return node
        else:
            if typemap.attrmap[data_type] is None:
                return node

    bpy_props = typemap.attrmap[data_type]
    for prop in bpy_props:
        parse_prop(options, node, data, prop, parent, default=bpy_props[prop])

    if options.include_custom_props:
        try:
            custom_props = [
                key for key in data.keys() if key not in data.bl_rna.properties.keys()
            ]
        except TypeError:
            pass
        else:
            for prop in custom_props:
                parse_custom_prop(options, node, data, prop)
    return node


def parse_path(options, rna_path):
    tokens = rna_path.split(".")
    head = tokens.pop(0)
    if head == "bpy":
        root = {}
        id = tokens.pop(0)
        root[id] = {}
        node = root[id]

        for token in tokens:
            node[token] = {}
            node = node[token]

        relative_path = ".".join(tokens)
        attr = getattr(bpy, id).path_resolve(relative_path)

        parse_node(options, node, attr, None)
        return root
    else:
        raise Exception("Invalid RNA path: %s" % rna_path)
