# SPDX-FileCopyrightText: 2023-2024 oeykmiih
# SPDX-License-Identifier: GPL-3.0-or-later
import logging

import bpy

from . import constructors


def try_setattr(obj, key, value):
    try:
        setattr(obj, key, value)
    except Exception:
        logging.exception("%s [ %s ] with the value: %s" % (obj, key, value))


def parse_custom_prop(options, node, data, prop):
    value = node[prop].pop("value")
    data[prop] = value
    data.id_properties_ensure()

    property = data.id_properties_ui(prop)
    property.update(**node[prop])


def parse_prop(options, node, data, prop):
    rna_props = data.bl_rna.properties
    prop_rna = rna_props[prop]

    match type(prop_rna.rna_type):
        case bpy.types.BoolProperty:
            if not prop_rna.is_readonly:
                try_setattr(data, prop, node[prop])
        case bpy.types.StringProperty:
            if not prop_rna.is_readonly:
                try_setattr(data, prop, node[prop])
        case bpy.types.FloatProperty | bpy.types.IntProperty:
            if not prop_rna.is_readonly:
                try_setattr(data, prop, node[prop])
        case bpy.types.EnumProperty:
            if not prop_rna.is_readonly:
                if prop_rna.is_enum_flag:
                    try_setattr(data, prop, set(node[prop]))
                else:
                    try_setattr(data, prop, node[prop])
        case bpy.types.CollectionProperty:
            subdata = getattr(data, prop)
            subnode = node[prop]

            fixed_type = type(prop_rna.fixed_type)

            # NOTE: For now it only supports CollectionProperty's with integer indexes
            # TODO: Handle string keys as well.

            # TODO: Expose options to import that can be used in different combinations:
            ## - is_update: should only update existing data
            ## - is_add: should only create new data
            ## - is_remove: should only remove data

            if len(subdata) < len(subnode):
                if fixed_type in constructors.constructors:
                    constructor = constructors.constructors[fixed_type]
                else:
                    try:
                        subdata.add()
                    except Exception:
                        raise NotImplementedError(
                            "This type has not been implemented: %s. %s in %s"
                            % (fixed_type, prop, data)
                        )
                    else:
                        subdata.remove(len(subdata) - 1)
                        constructor = constructors.default

                for i in range(len(subdata), len(subnode)):
                    constructor(subdata, subnode[i])
            elif len(subdata) > len(subnode):
                pass

            for i, subnode_item in enumerate(subnode):
                parse_node(options, subnode_item, subdata[i])

            subdata.update()
        case bpy.types.PointerProperty:
            subdata = getattr(data, prop)
            fixed_type = type(prop_rna.fixed_type)

            parse_node(options, node[prop], subdata)

            if fixed_type in {bpy.types.CurveMapping}:
                subdata.update()


def parse_node(options, node, data) -> None:
    # NOTE: This tries to safeguard against values that directly require other values
    ## (1st depth level), will still throw an error at:
    ## - higher depth
    ## - when available values are not updated

    rna_props = data.bl_rna.properties.keys()

    bpy_props = []
    custom_props = []
    for key in node:
        if key not in rna_props:
            custom_props.append(key)
        else:
            bpy_props.append(key)

    if options.include_custom_props:
        for key in custom_props:
            parse_custom_prop(options, node, data, key)

    again = []
    for key in bpy_props:
        try:
            parse_prop(options, node, data, key)
        except TypeError:
            again.append(key)

    if len(again):
        for key in again:
            parse_prop(options, node, data, key)


def parse_root(options, root: str) -> None:
    for key in root.keys():
        attr = getattr(bpy, key)
        node = root[key]

        parse_node(options, node, attr)
